# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=wrong-import-position

import timeit
# Log the start time
start_time = timeit.default_timer()

import sys
import uuid

import azure.cli.core.telemetry as telemetry
from azure.cli.core import get_default_cli
from knack.completion import ARGCOMPLETE_ENV_NAME
from knack.log import get_logger

__author__ = "Microsoft Corporation <python@microsoft.com>"
__version__ = "2.27.2.post20210825155733"


# A workaround for https://bugs.python.org/issue32502 (https://github.com/Azure/azure-cli/issues/5184)
# If uuid1 raises ValueError, use uuid4 instead.
try:
    uuid.uuid1()
except ValueError:
    uuid.uuid1 = uuid.uuid4


logger = get_logger(__name__)


def cli_main(cli, args):
    return cli.invoke(args)


az_cli = get_default_cli()

telemetry.set_application(az_cli, ARGCOMPLETE_ENV_NAME)

# Log the init finish time
init_finish_time = timeit.default_timer()

try:
    telemetry.start()

    exit_code = cli_main(az_cli, sys.argv[1:])

    if exit_code == 0:
        telemetry.set_success()

    sys.exit(exit_code)

except KeyboardInterrupt:
    telemetry.set_user_fault('Keyboard interrupt is captured.')
    sys.exit(1)
except SystemExit as ex:  # some code directly call sys.exit, this is to make sure command metadata is logged
    exit_code = ex.code if ex.code is not None else 1
    raise ex

finally:
    try:
        # Log the invoke finish time
        invoke_finish_time = timeit.default_timer()
        logger.info("Command ran in %.3f seconds (init: %.3f, invoke: %.3f)",
                    invoke_finish_time - start_time,
                    init_finish_time - start_time,
                    invoke_finish_time - init_finish_time)
    except NameError:
        pass

    try:
        # check for new version auto-upgrade
        if exit_code == 0 and az_cli.config.getboolean('auto-upgrade', 'enable', False) and \
                sys.argv[1] != 'upgrade' and (sys.argv[1] != 'extension' or sys.argv[2] != 'update'):
            from azure.cli.core._session import VERSIONS  # pylint: disable=ungrouped-imports
            from azure.cli.core.util import get_cached_latest_versions, _VERSION_UPDATE_TIME  # pylint: disable=ungrouped-imports
            if VERSIONS[_VERSION_UPDATE_TIME]:
                import datetime
                version_update_time = datetime.datetime.strptime(VERSIONS[_VERSION_UPDATE_TIME], '%Y-%m-%d %H:%M:%S.%f')
                if datetime.datetime.now() > version_update_time + datetime.timedelta(days=11):
                    get_cached_latest_versions()
                from packaging.version import parse
                if parse(VERSIONS['versions']['core']['local']) < parse(VERSIONS['versions']['core']['pypi']):  # pylint: disable=line-too-long
                    import subprocess
                    import platform
                    from azure.cli.core.azclierror import UnclassifiedUserFault  # pylint: disable=ungrouped-imports
                    logger.warning("New Azure CLI version available. Running 'az upgrade' to update automatically.")
                    update_all = az_cli.config.getboolean('auto-upgrade', 'all', True)
                    prompt = az_cli.config.getboolean('auto-upgrade', 'prompt', True)
                    cmd = ['az', 'upgrade', '--all', str(update_all)]
                    if prompt:
                        from knack.prompting import verify_is_a_tty, NoTTYException  # pylint: disable=ungrouped-imports
                        az_upgrade_run = True
                        try:
                            verify_is_a_tty()
                        except NoTTYException:
                            az_upgrade_run = False
                            err_msg = "Unable to prompt for auto upgrade as no tty available. " \
                                "Run 'az config set auto-upgrade.prompt=no' to allow auto upgrade with no prompt."
                            logger.warning(err_msg)
                            telemetry.set_exception(UnclassifiedUserFault(err_msg), fault_type='auto-upgrade-failed')
                        else:
                            upgrade_exit_code = subprocess.call(cmd, shell=platform.system() == 'Windows')
                    else:
                        import os
                        devnull = open(os.devnull, 'w')
                        cmd.append('-y')
                        upgrade_exit_code = subprocess.call(cmd, shell=platform.system() == 'Windows', stdout=devnull)
                    if az_upgrade_run and upgrade_exit_code != 0:
                        err_msg = "Auto upgrade failed with exit code {}".format(exit_code)
                        logger.warning(err_msg)
                        telemetry.set_exception(UnclassifiedUserFault(err_msg), fault_type='auto-upgrade-failed')
    except IndexError:
        pass
    except Exception as ex:  # pylint: disable=broad-except
        logger.warning("Auto upgrade failed. %s", str(ex))
        telemetry.set_exception(ex, fault_type='auto-upgrade-failed')

    telemetry.set_init_time_elapsed("{:.6f}".format(init_finish_time - start_time))
    telemetry.set_invoke_time_elapsed("{:.6f}".format(invoke_finish_time - init_finish_time))
    telemetry.conclude()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
import os
import re

from azure.cli.core.commands.arm import ArmTemplateBuilder

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse  # pylint: disable=import-error

from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)


MSI_LOCAL_ID = '[system]'


def get_target_network_api(cli_ctx):
    """ Since most compute calls don't need advanced network functionality, we can target a supported, but not
        necessarily latest, network API version is order to avoid having to re-record every test that uses VM create
        (which there are a lot) whenever NRP bumps their API version (which is often)!
    """
    from azure.cli.core.profiles import get_api_version, ResourceType, AD_HOC_API_VERSIONS
    version = get_api_version(cli_ctx, ResourceType.MGMT_NETWORK)
    if cli_ctx.cloud.profile == 'latest':
        version = AD_HOC_API_VERSIONS[ResourceType.MGMT_NETWORK]['vm_default_target_network']
    return version


def read_content_if_is_file(string_or_file):
    content = string_or_file
    if os.path.exists(string_or_file):
        with open(string_or_file, 'r') as f:
            content = f.read()
    return content


def _resolve_api_version(cli_ctx, provider_namespace, resource_type, parent_path):
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.cli.core.profiles import ResourceType
    client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES)
    provider = client.providers.get(provider_namespace)

    # If available, we will use parent resource's api-version
    resource_type_str = (parent_path.split('/')[0] if parent_path else resource_type)

    rt = [t for t in provider.resource_types  # pylint: disable=no-member
          if t.resource_type.lower() == resource_type_str.lower()]
    if not rt:
        raise CLIError('Resource type {} not found.'.format(resource_type_str))
    if len(rt) == 1 and rt[0].api_versions:
        npv = [v for v in rt[0].api_versions if 'preview' not in v.lower()]
        return npv[0] if npv else rt[0].api_versions[0]
    raise CLIError(
        'API version is required and could not be resolved for resource {}'
        .format(resource_type))


def log_pprint_template(template):
    logger.info('==== BEGIN TEMPLATE ====')
    logger.info(json.dumps(template, indent=2))
    logger.info('==== END TEMPLATE ====')


def check_existence(cli_ctx, value, resource_group, provider_namespace, resource_type,
                    parent_name=None, parent_type=None):
    # check for name or ID and set the type flags
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.core.exceptions import HttpResponseError
    from msrestazure.tools import parse_resource_id
    from azure.cli.core.profiles import ResourceType
    id_parts = parse_resource_id(value)
    resource_client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES,
                                              subscription_id=id_parts.get('subscription', None)).resources
    rg = id_parts.get('resource_group', resource_group)
    ns = id_parts.get('namespace', provider_namespace)

    if parent_name and parent_type:
        parent_path = '{}/{}'.format(parent_type, parent_name)
        resource_name = id_parts.get('child_name_1', value)
        resource_type = id_parts.get('child_type_1', resource_type)
    else:
        parent_path = ''
        resource_name = id_parts['name']
        resource_type = id_parts.get('type', resource_type)
    api_version = _resolve_api_version(cli_ctx, provider_namespace, resource_type, parent_path)

    try:
        resource_client.get(rg, ns, parent_path, resource_type, resource_name, api_version)
        return True
    except HttpResponseError:
        return False


def create_keyvault_data_plane_client(cli_ctx):
    from azure.cli.command_modules.keyvault._client_factory import keyvault_data_plane_factory
    return keyvault_data_plane_factory(cli_ctx)


def get_key_vault_base_url(cli_ctx, vault_name):
    suffix = cli_ctx.cloud.suffixes.keyvault_dns
    return 'https://{}{}'.format(vault_name, suffix)


def list_sku_info(cli_ctx, location=None):
    from ._client_factory import _compute_client_factory

    def _match_location(loc, locations):
        return next((x for x in locations if x.lower() == loc.lower()), None)

    client = _compute_client_factory(cli_ctx)
    result = client.resource_skus.list()
    if location:
        result = [r for r in result if _match_location(location, r.locations)]
    return result


# pylint: disable=too-many-statements
def normalize_disk_info(image_data_disks=None,
                        data_disk_sizes_gb=None, attach_data_disks=None, storage_sku=None,
                        os_disk_caching=None, data_disk_cachings=None, size='', ephemeral_os_disk=False,
                        data_disk_delete_option=None):
    from msrestazure.tools import is_valid_resource_id
    from ._validators import validate_delete_options
    is_lv_size = re.search('_L[0-9]+s', size, re.I)
    # we should return a dictionary with info like below
    # {
    #   'os': { caching: 'Read', write_accelerator: None},
    #   0: { caching: 'None', write_accelerator: True},
    #   1: { caching: 'None', write_accelerator: True},
    # }
    info = {}
    used_luns = set()

    attach_data_disks = attach_data_disks or []
    data_disk_sizes_gb = data_disk_sizes_gb or []
    image_data_disks = image_data_disks or []

    data_disk_delete_option = validate_delete_options(attach_data_disks, data_disk_delete_option)
    info['os'] = {}
    # update os diff disk settings
    if ephemeral_os_disk:
        info['os']['diffDiskSettings'] = {'option': 'Local'}
        # local os disks require readonly caching, default to ReadOnly if os_disk_caching not specified.
        if not os_disk_caching:
            os_disk_caching = 'ReadOnly'

    # add managed image data disks
    for data_disk in image_data_disks:
        i = data_disk['lun']
        info[i] = {
            'lun': i,
            'managedDisk': {'storageAccountType': None},
            'createOption': 'fromImage'
        }
        used_luns.add(i)

    # add empty data disks, do not use existing luns
    i = 0
    sizes_copy = list(data_disk_sizes_gb)
    while sizes_copy:
        # get free lun
        while i in used_luns:
            i += 1

        used_luns.add(i)

        info[i] = {
            'lun': i,
            'managedDisk': {'storageAccountType': None},
            'createOption': 'empty',
            'diskSizeGB': sizes_copy.pop(0),
            'deleteOption': data_disk_delete_option if isinstance(data_disk_delete_option, str) else None
        }

    # update storage skus for managed data disks
    if storage_sku is not None:
        update_disk_sku_info(info, storage_sku)

    # check that os storage account type is not UltraSSD_LRS
    if info['os'].get('storageAccountType', "").lower() == 'ultrassd_lrs':
        logger.warning("Managed os disk storage account sku cannot be UltraSSD_LRS. Using service default.")
        info['os']['storageAccountType'] = None

    # add attached data disks
    i = 0
    attach_data_disks_copy = list(attach_data_disks)
    while attach_data_disks_copy:
        # get free lun
        while i in used_luns:
            i += 1

        used_luns.add(i)

        # use free lun
        info[i] = {
            'lun': i,
            'createOption': 'attach'
        }

        d = attach_data_disks_copy.pop(0)

        if is_valid_resource_id(d):
            info[i]['managedDisk'] = {'id': d}
            if data_disk_delete_option:
                info[i]['deleteOption'] = data_disk_delete_option if isinstance(data_disk_delete_option, str) \
                    else data_disk_delete_option.get(info[i]['name'], None)
        else:
            info[i]['vhd'] = {'uri': d}
            info[i]['name'] = d.split('/')[-1].split('.')[0]
            if data_disk_delete_option:
                info[i]['deleteOption'] = data_disk_delete_option if isinstance(data_disk_delete_option, str) \
                    else data_disk_delete_option.get(info[i]['name'], None)

    # fill in data disk caching
    if data_disk_cachings:
        update_disk_caching(info, data_disk_cachings)

    # default os disk caching to 'ReadWrite' unless set otherwise
    if os_disk_caching:
        info['os']['caching'] = os_disk_caching
    else:
        info['os']['caching'] = 'None' if is_lv_size else 'ReadWrite'

    # error out on invalid vm sizes
    if is_lv_size:
        for v in info.values():
            if v.get('caching', 'None').lower() != 'none':
                raise CLIError('usage error: for Lv series of machines, "None" is the only supported caching mode')

    result_info = {'os': info['os']}

    # in python 3 insertion order matters during iteration. This ensures that luns are retrieved in numerical order
    for key in sorted([key for key in info if key != 'os']):
        result_info[key] = info[key]

    return result_info


def update_disk_caching(model, caching_settings):

    def _update(model, lun, value):
        if isinstance(model, dict):
            luns = model.keys() if lun is None else [lun]
            for lun_item in luns:
                if lun_item not in model:
                    raise CLIError("Data disk with lun of '{}' doesn't exist. Existing luns: {}."
                                   .format(lun_item, list(model.keys())))
                model[lun_item]['caching'] = value
        else:
            if lun is None:
                disks = [model.os_disk] + (model.data_disks or [])
            elif lun == 'os':
                disks = [model.os_disk]
            else:
                disk = next((d for d in model.data_disks if d.lun == lun), None)
                if not disk:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                disks = [disk]
            for disk in disks:
                disk.caching = value

    if len(caching_settings) == 1 and '=' not in caching_settings[0]:
        _update(model, None, caching_settings[0])
    else:
        for x in caching_settings:
            if '=' not in x:
                raise CLIError("usage error: please use 'LUN=VALUE' to configure caching on individual disk")
            lun, value = x.split('=', 1)
            lun = lun.lower()
            lun = int(lun) if lun != 'os' else lun
            _update(model, lun, value)


def update_write_accelerator_settings(model, write_accelerator_settings):

    def _update(model, lun, value):
        if isinstance(model, dict):
            luns = model.keys() if lun is None else [lun]
            for lun_item in luns:
                if lun_item not in model:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun_item))
                model[lun_item]['writeAcceleratorEnabled'] = value
        else:
            if lun is None:
                disks = [model.os_disk] + (model.data_disks or [])
            elif lun == 'os':
                disks = [model.os_disk]
            else:
                disk = next((d for d in model.data_disks if d.lun == lun), None)
                if not disk:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                disks = [disk]
            for disk in disks:
                disk.write_accelerator_enabled = value

    if len(write_accelerator_settings) == 1 and '=' not in write_accelerator_settings[0]:
        _update(model, None, write_accelerator_settings[0].lower() == 'true')
    else:
        for x in write_accelerator_settings:
            if '=' not in x:
                raise CLIError("usage error: please use 'LUN=VALUE' to configure write accelerator"
                               " on individual disk")
            lun, value = x.split('=', 1)
            lun = lun.lower()
            lun = int(lun) if lun != 'os' else lun
            _update(model, lun, value.lower() == 'true')


def get_storage_blob_uri(cli_ctx, storage):
    from azure.cli.core.profiles._shared import ResourceType
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    if urlparse(storage).scheme:
        storage_uri = storage
    else:
        storage_mgmt_client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_STORAGE)
        storage_accounts = storage_mgmt_client.storage_accounts.list()
        storage_account = next((a for a in list(storage_accounts)
                                if a.name.lower() == storage.lower()), None)
        if storage_account is None:
            raise CLIError('{} does\'t exist.'.format(storage))
        storage_uri = storage_account.primary_endpoints.blob
    return storage_uri


def update_disk_sku_info(info_dict, skus):
    usage_msg = 'Usage:\n\t[--storage-sku SKU | --storage-sku ID=SKU ID=SKU ID=SKU...]\n' \
                'where each ID is "os" or a 0-indexed lun.'

    def _update(info, lun, value):
        luns = info.keys()
        if lun not in luns:
            raise CLIError("Data disk with lun of '{}' doesn't exist. Existing luns: {}.".format(lun, luns))
        if lun == 'os':
            info[lun]['storageAccountType'] = value
        else:
            info[lun]['managedDisk']['storageAccountType'] = value

    if len(skus) == 1 and '=' not in skus[0]:
        for lun in info_dict.keys():
            _update(info_dict, lun, skus[0])
    else:
        for sku in skus:
            if '=' not in sku:
                raise CLIError("A sku's format is incorrect.\n{}".format(usage_msg))

            lun, value = sku.split('=', 1)
            lun = lun.lower()
            try:
                lun = int(lun) if lun != "os" else lun
            except ValueError:
                raise CLIError("A sku ID is incorrect.\n{}".format(usage_msg))
            _update(info_dict, lun, value)


def is_shared_gallery_image_id(image_reference):
    if not image_reference:
        return False

    shared_gallery_id_pattern = re.compile(r'^/SharedGalleries/[^/]*/Images/[^/]*/Versions/.*$', re.IGNORECASE)
    if shared_gallery_id_pattern.match(image_reference):
        return True

    return False


class ArmTemplateBuilder20190401(ArmTemplateBuilder):

    def __init__(self):
        super().__init__()
        self.template['$schema'] = 'https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#'



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long

from knack.arguments import CLIArgumentType


def load_arguments(self, _):
    from azure.mgmt.redis.models import RebootType, RedisKeyType, SkuName, TlsVersion, ReplicationRole
    from azure.cli.command_modules.redis._validators import JsonString, ScheduleEntryList
    from azure.cli.command_modules.redis.custom import allowed_c_family_sizes, allowed_p_family_sizes
    from azure.cli.core.commands.parameters import get_enum_type, tags_type, zones_type
    from azure.cli.core.commands.parameters import get_resource_name_completion_list

    with self.argument_context('redis') as c:
        cache_name = CLIArgumentType(options_list=['--name', '-n'], help='Name of the Redis cache.', id_part='name',
                                     completer=get_resource_name_completion_list('Microsoft.Cache/redis'))
        format_type = CLIArgumentType(options_list=['--file-format'], help='Format of the blob (Currently rdb is the only supported format, with other formats expected in the future)')

        c.argument('name', arg_type=cache_name)
        c.argument('redis_configuration', help='JSON encoded configuration settings. Use @{file} to load from a file.',
                   type=JsonString)
        c.argument('reboot_type', arg_type=get_enum_type(RebootType), help='Which Redis node(s) to reboot. Depending on this value data loss is possible.')
        c.argument('key_type', arg_type=get_enum_type(RedisKeyType), help='The Redis access key to regenerate.')
        c.argument('files', help='SAS url for blobs that needs to be imported', nargs='+')
        c.argument('format', arg_type=format_type)
        c.argument('file_format', arg_type=format_type)
        c.argument('container', help='SAS url for container where data needs to be exported to')
        c.argument('prefix', help='Prefix to use for exported files')
        c.argument('cache_name', arg_type=cache_name)
        c.argument('shard_count', type=int, help='The number of shards to be created on a Premium Cluster Cache.')
        c.argument('subnet_id', help='The full resource ID of a subnet in a virtual network to deploy the redis cache in. Example format /subscriptions/{subid}/resourceGroups/{resourceGroupName}/providers/Microsoft.{Network|ClassicNetwork}/virtualNetworks/vnet1/subnets/subnet1')
        c.argument('static_ip', help='Specify a static ip if required for the VNET. If you do not specify a static IP then an IP address is chosen automatically')
        c.argument('tenant_settings', arg_type=tags_type, help='Space-separated tenant settings in key[=value] format')
        c.argument('tags', arg_type=tags_type)
        c.argument('zones', arg_type=zones_type)
        c.argument('shard_id', type=int)
        c.argument('sku', help='Type of Redis cache.', arg_type=get_enum_type(SkuName))
        c.argument('minimum_tls_version', help='Specifies the TLS version required by clients to connect to cache', arg_type=get_enum_type(TlsVersion))
        c.argument('vm_size', arg_type=get_enum_type(allowed_c_family_sizes + allowed_p_family_sizes), help='Size of Redis cache to deploy. Basic and Standard Cache sizes start with C. Premium Cache sizes start with P')
        c.argument('enable_non_ssl_port', action='store_true', help='If specified, then the non-ssl redis server port (6379) will be enabled.')
        c.argument('replicas_per_master', help='The number of replicas to be created per master.')
        c.argument('redis_version', help='Redis version. Only major version will be used in create/update request with current valid values: (4, 6)')

    with self.argument_context('redis firewall-rules list') as c:
        c.argument('cache_name', arg_type=cache_name, id_part=None)
        c.argument('rule_name', help='Name of the firewall rule')

    with self.argument_context('redis firewall-rules') as c:
        c.argument('end_ip', help='Highest IP address included in the range.')
        c.argument('rule_name', help='The name of the firewall rule.')
        c.argument('start_ip', help='Lowest IP address included in the range.')

    with self.argument_context('redis force-reboot') as c:
        c.argument('shard_id', help='If clustering is enabled, the ID of the shard to be rebooted.')

    with self.argument_context('redis server-link') as c:
        c.argument('name', arg_type=cache_name, id_part=None)
        c.argument('server_to_link', help='Resource ID or name of the redis cache to be linked')
        c.argument('replication_role', help='Role of the redis cache to be linked', arg_type=get_enum_type(ReplicationRole))
        c.argument('linked_server_name', help='Name of the linked redis cache')

    with self.argument_context('redis patch-schedule') as c:
        c.argument('name', arg_type=cache_name, id_part=None)
        c.argument('schedule_entries', help="List of Patch schedule entries. Example Value:[{\"dayOfWeek\":\"Monday\",\"startHourUtc\":\"00\",\"maintenanceWindow\":\"PT5H\"}]", type=ScheduleEntryList)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long

from knack.util import CLIError
from msrest.exceptions import ValidationError  # pylint: disable=import-error
from azure.core.exceptions import HttpResponseError


def netappfiles_exception_handler(ex):
    if isinstance(ex, (HttpResponseError, ValidationError, ValueError)):
        message = ex
        raise CLIError(message)

    import sys

    from six import reraise
    reraise(*sys.exc_info())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from .MetricAlertConditionListener import MetricAlertConditionListener


op_conversion = {
    '=': 'Equals',
    '!=': 'NotEquals',
    '>': 'GreaterThan',
    '>=': 'GreaterThanOrEqual',
    '<': 'LessThan',
    '<=': 'LessThanOrEqual',
    '><': 'GreaterOrLessThan'
}

agg_conversion = {
    'avg': 'Average',
    'min': 'Minimum',
    'max': 'Maximum',
    'total': 'Total',
    'count': 'Count'
}

sens_conversion = {
    'low': 'Low',
    'medium': 'Medium',
    'high': 'High',
}

dim_op_conversion = {
    'includes': 'Include',
    'excludes': 'Exclude'
}


# This class defines a complete listener for a parse tree produced by MetricAlertConditionParser.
class MetricAlertConditionValidator(MetricAlertConditionListener):

    def __init__(self):
        super(MetricAlertConditionValidator, self).__init__()
        self.parameters = {}
        self._dimension_index = 0

    # Exit a parse tree produced by MetricAlertConditionParser#aggregation.
    def exitAggregation(self, ctx):
        aggregation = agg_conversion[ctx.getText().strip()]
        self.parameters['time_aggregation'] = aggregation

    # Exit a parse tree produced by MetricAlertConditionParser#namespace.
    def exitNamespace(self, ctx):
        self.parameters['metric_namespace'] = ctx.getText().strip()

    # Exit a parse tree produced by MetricAlertConditionParser#metric.
    def exitMetric(self, ctx):
        self.parameters['metric_name'] = ctx.getText().strip()

    # Exit a parse tree produced by MetricAlertConditionParser#operator.
    def exitOperator(self, ctx):
        self.parameters['operator'] = op_conversion[ctx.getText().strip()]

    # Exit a parse tree produced by MetricAlertConditionParser#threshold.
    def exitThreshold(self, ctx):
        self.parameters['threshold'] = ctx.getText().strip()

    def exitDynamic(self, ctx):
        self.parameters['failing_periods'] = {}

    def exitDyn_sensitivity(self, ctx):
        sensitivity = sens_conversion[ctx.getText().strip().lower()]
        self.parameters['alert_sensitivity'] = sensitivity

    def exitDyn_violations(self, ctx):
        min_failing_periods_to_alert = float(ctx.getText().strip())
        if min_failing_periods_to_alert < 1 or min_failing_periods_to_alert > 6:
            message = "Violations {} should be 1-6."
            raise ValueError(message.format(min_failing_periods_to_alert))
        self.parameters['failing_periods']['min_failing_periods_to_alert'] = min_failing_periods_to_alert

    def exitDyn_windows(self, ctx):
        number_of_evaluation_periods = float(ctx.getText().strip())
        min_failing_periods_to_alert = self.parameters['failing_periods']['min_failing_periods_to_alert']
        if number_of_evaluation_periods < 1 or number_of_evaluation_periods > 6:
            message = "Windows {} should be 1-6."
            raise ValueError(message.format(number_of_evaluation_periods))
        if min_failing_periods_to_alert > number_of_evaluation_periods:
            message = "Violations {} should be smaller or equal to windows {}."
            raise ValueError(message.format(min_failing_periods_to_alert, number_of_evaluation_periods))
        self.parameters['failing_periods']['number_of_evaluation_periods'] = number_of_evaluation_periods

    def exitDyn_datetime(self, ctx):
        from msrest.serialization import Deserializer
        from msrest.exceptions import DeserializationError
        datetime_str = ctx.getText().strip()
        try:
            self.parameters['ignore_data_before'] = Deserializer.deserialize_iso(datetime_str)
        except DeserializationError:
            message = "Datetime {} is not a valid ISO-8601 format"
            raise ValueError(message.format(datetime_str))

    # Enter a parse tree produced by MetricAlertConditionParser#dimensions.
    def enterDimensions(self, ctx):
        self.parameters['dimensions'] = []

    # Enter a parse tree produced by MetricAlertConditionParser#dimension.
    def enterDimension(self, ctx):
        self.parameters['dimensions'].append({})

    # Exit a parse tree produced by MetricAlertConditionParser#dimension.
    def exitDimension(self, ctx):
        self._dimension_index = self._dimension_index + 1

    # Exit a parse tree produced by MetricAlertConditionParser#dname.
    def exitDim_name(self, ctx):
        self.parameters['dimensions'][self._dimension_index]['name'] = ctx.getText().strip()

    # Exit a parse tree produced by MetricAlertConditionParser#dop.
    def exitDim_operator(self, ctx):
        op_text = ctx.getText().strip()
        self.parameters['dimensions'][self._dimension_index]['operator'] = dim_op_conversion[op_text.lower()]

    # Exit a parse tree produced by MetricAlertConditionParser#dvalues.
    def exitDim_values(self, ctx):
        dvalues = ctx.getText().strip().split(' ')
        self.parameters['dimensions'][self._dimension_index]['values'] = [x for x in dvalues if x not in ['', 'or']]

    def result(self):
        from azure.mgmt.monitor.models import MetricCriteria, MetricDimension, DynamicMetricCriteria, \
            DynamicThresholdFailingPeriods
        dim_params = self.parameters.get('dimensions', [])
        dimensions = []
        for dim in dim_params:
            dimensions.append(MetricDimension(**dim))
        self.parameters['dimensions'] = dimensions
        self.parameters['name'] = ''  # will be auto-populated later

        if 'failing_periods' in self.parameters:
            # dynamic metric criteria
            failing_periods = DynamicThresholdFailingPeriods(**self.parameters['failing_periods'])
            self.parameters['failing_periods'] = failing_periods
            return DynamicMetricCriteria(**self.parameters)

        # static metric criteria
        return MetricCriteria(**self.parameters)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


# pylint: disable=unused-argument, line-too-long
def create_subscription_diagnostic_settings(cmd, client,
                                            name, logs,
                                            location,
                                            event_hub_name=None,
                                            event_hub_auth_rule=None,
                                            storage_account=None,
                                            service_bus_rule=None,
                                            workspace=None):
    from azure.mgmt.monitor.models import SubscriptionDiagnosticSettingsResource
    from azure.cli.core.commands.client_factory import get_subscription_id
    parameters = SubscriptionDiagnosticSettingsResource(storage_account_id=storage_account,
                                                        workspace_id=workspace,
                                                        event_hub_name=event_hub_name,
                                                        event_hub_authorization_rule_id=event_hub_auth_rule,
                                                        service_bus_rule_id=service_bus_rule,
                                                        logs=logs,
                                                        location=location)
    return client.create_or_update(subscription_id=get_subscription_id(cmd.cli_ctx), parameters=parameters, name=name)


def update_subscription_diagnostic_settings(cmd, instance, event_hub_name=None, logs=None,
                                            event_hub_auth_rule=None, storage_account=None,
                                            service_bus_rule=None, workspace=None):
    with cmd.update_context(instance) as c:
        c.set_param('storage_account_id', storage_account)
        c.set_param('workspace_id', workspace)
        c.set_param('event_hub_name', event_hub_name)
        c.set_param('event_hub_authorization_rule_id', event_hub_auth_rule)
        c.set_param('service_bus_rule_id', service_bus_rule)
        c.set_param('logs', logs)
    return instance



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import argparse

from knack.log import get_logger
from knack.util import CLIError

from azure.mgmt.cosmosdb.models import (
    Location,
    DatabaseRestoreResource
)

logger = get_logger(__name__)


# pylint: disable=protected-access, too-few-public-methods
class CreateLocation(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.locations is None:
            namespace._deprecated_location_format = False
            namespace.locations = []

        if any("regionname" in s.lower() for s in values):
            keys_found = set()
            _name = ""
            _failover = 0
            _is_zr = False
            for item in values:
                kvp = item.split('=', 1)
                _key = kvp[0].lower()
                if _key in keys_found:
                    # pylint: disable=line-too-long
                    raise CLIError('usage error: --locations [KEY=VALUE ...]. KEY {} provided multiple times'.format(_key))
                keys_found.add(_key)
                if _key == "regionname":
                    _name = kvp[1]
                elif _key == "failoverpriority":
                    _failover = int(kvp[1])
                elif _key == "iszoneredundant":
                    _is_zr = kvp[1].lower() == "true"
                else:
                    # pylint: disable=line-too-long
                    raise CLIError('usage error: --locations [KEY=VALUE ...]. KEY {} not supported. Valid keys are regionName, failoverPriority, and isZoneRedundant'.format(_key))
            namespace.locations.append(
                Location(location_name=_name,
                         failover_priority=_failover,
                         is_zone_redundant=_is_zr))
        else:
            # pylint: disable=line-too-long
            if not namespace._deprecated_location_format:
                logger.warning('The regionName=failoverPriority method of specifying locations is deprecated. Use --locations KEY=VALUE [KEY=VALUE ...] to specify the regionName, failoverPriority, and isZoneRedundant properties of the location. Multiple locations can be specified by including more than one --locations argument.')
            namespace._deprecated_location_format = True

            for item in values:
                comps = item.split('=', 1)
                namespace.locations.append(
                    Location(location_name=comps[0],
                             failover_priority=int(comps[1]),
                             is_zone_redundant=False))


class CreateDatabaseRestoreResource(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.databases_to_restore is None:
            namespace.databases_to_restore = []
        if not values:
            # pylint: disable=line-too-long
            raise CLIError('usage error: --databases-to-restore [name=DatabaseName collections=CollectionName1 CollectionName2 ...]')
        database_restore_resource = DatabaseRestoreResource()
        i = 0
        for item in values:
            if i == 0:
                kvp = item.split('=', 1)
                if len(kvp) != 2 or kvp[0].lower() != 'name':
                    # pylint: disable=line-too-long
                    raise CLIError('usage error: --databases-to-restore [name=DatabaseName collections=CollectionName1 CollectionName2 ...]')
                database_name = kvp[1]
                database_restore_resource.database_name = database_name
            elif i == 1:
                kvp = item.split('=', 1)
                if len(kvp) != 2 or kvp[0].lower() != 'collections':
                    # pylint: disable=line-too-long
                    raise CLIError('usage error: --databases-to-restore [name=DatabaseName collections=CollectionName1 CollectionName2 ...]')
                database_restore_resource.collection_names = []
                collection_name = kvp[1]
                database_restore_resource.collection_names.append(collection_name)
            else:
                if database_restore_resource.collection_names is None:
                    database_restore_resource.collection_names = []
                database_restore_resource.collection_names.append(item)
            i += 1
        namespace.databases_to_restore.append(database_restore_resource)


class UtcDatetimeAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        """ Parse a date value and return the ISO8601 string. """
        import dateutil.parser
        import dateutil.tz

        accepted_formats = []
        accepted_formats.append('date (yyyy-mm-dd)')
        accepted_formats.append('time (hh:mm:ss.xxxxx)')
        accepted_formats.append('timezone (+/-hh:mm)')
        help_string = 'Format: ' + ' '.join(accepted_formats)
        value_string = ''.join(values)
        dt_val = None
        try:
            # attempt to parse ISO 8601
            dt_val = dateutil.parser.parse(value_string)
        except ValueError:
            pass

        if not dt_val:
            raise CLIError("Unable to parse: '{}'. Expected format: {}".format(value_string, help_string))

        if not dt_val.tzinfo:
            dt_val = dt_val.replace(tzinfo=dateutil.tz.tzutc())

        iso_string = dt_val.isoformat()
        setattr(namespace, self.dest, iso_string)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def _build_frontend_ip_config(cmd, name, public_ip_id=None, subnet_id=None, private_ip_address=None,
                              private_ip_allocation=None, zone=None, private_ip_address_version=None,
                              enable_private_link=False,
                              private_link_configuration_id=None):
    frontend_ip_config = {
        'name': name
    }

    if public_ip_id:
        frontend_ip_config.update({
            'properties': {
                'publicIPAddress': {'id': public_ip_id}
            }
        })
    else:
        frontend_ip_config.update({
            'properties': {
                'privateIPAllocationMethod': private_ip_allocation,
                'privateIPAddress': private_ip_address,
                'subnet': {'id': subnet_id}
            }
        })

    if zone and cmd.supported_api_version(min_api='2017-06-01'):
        frontend_ip_config['zones'] = zone

    if private_ip_address_version and cmd.supported_api_version(min_api='2019-04-01'):
        frontend_ip_config['properties']['privateIPAddressVersion'] = private_ip_address_version

    if enable_private_link is True and cmd.supported_api_version(min_api='2020-05-01'):
        frontend_ip_config['properties'].update({
            'privateLinkConfiguration': {'id': private_link_configuration_id}
        })

    return frontend_ip_config


def _build_appgw_private_link_ip_configuration(name,
                                               private_link_ip_address,
                                               private_link_ip_allocation_method,
                                               private_link_primary,
                                               private_link_subnet_id):
    return {
        'name': name,
        'properties': {
            'privateIPAddress': private_link_ip_address,
            'privateIPAllocationMethod': private_link_ip_allocation_method,
            'primary': private_link_primary,
            'subnet': {'id': private_link_subnet_id}
        }
    }


# pylint: disable=too-many-locals, too-many-statements, too-many-branches
def build_application_gateway_resource(cmd, name, location, tags, sku_name, sku_tier, capacity, servers, frontend_port,
                                       private_ip_address, private_ip_allocation,
                                       cert_data, cert_password, key_vault_secret_id,
                                       cookie_based_affinity, http_settings_protocol, http_settings_port,
                                       http_listener_protocol, routing_rule_type, public_ip_id, subnet_id,
                                       connection_draining_timeout, enable_http2, min_capacity, zones,
                                       custom_error_pages, firewall_policy, max_capacity,
                                       user_assigned_identity,
                                       enable_private_link=False,
                                       private_link_name=None,
                                       private_link_ip_address=None,
                                       private_link_ip_allocation_method=None,
                                       private_link_primary=None,
                                       private_link_subnet_id=None,
                                       trusted_client_certificates=None,
                                       ssl_profile=None,
                                       ssl_profile_id=None,
                                       ssl_cert_name=None):

    # set the default names
    frontend_public_ip_name = 'appGatewayFrontendIP'
    frontend_private_ip_name = 'appGatewayPrivateFrontendIP'
    backend_pool_name = 'appGatewayBackendPool'
    frontend_port_name = 'appGatewayFrontendPort'
    http_listener_name = 'appGatewayHttpListener'
    http_settings_name = 'appGatewayBackendHttpSettings'
    routing_rule_name = 'rule1'

    if not ssl_cert_name:
        ssl_cert_name = '{}SslCert'.format(name)

    ssl_cert = None

    backend_address_pool = {'name': backend_pool_name}
    if servers:
        backend_address_pool['properties'] = {'BackendAddresses': servers}

    def _ag_subresource_id(_type, name):
        return "[concat(variables('appGwID'), '/{}/{}')]".format(_type, name)

    frontend_port_id = _ag_subresource_id('frontendPorts', frontend_port_name)
    http_listener_id = _ag_subresource_id('httpListeners', http_listener_name)
    backend_address_pool_id = _ag_subresource_id('backendAddressPools', backend_pool_name)
    backend_http_settings_id = _ag_subresource_id('backendHttpSettingsCollection',
                                                  http_settings_name)
    ssl_cert_id = _ag_subresource_id('sslCertificates', ssl_cert_name)

    private_link_configuration_id = None
    privateLinkConfigurations = []
    if cmd.supported_api_version(min_api='2020-05-01') and enable_private_link:
        private_link_configuration_id = _ag_subresource_id('privateLinkConfigurations',
                                                           private_link_name)

        default_private_link_ip_config = _build_appgw_private_link_ip_configuration(
            'PrivateLinkDefaultIPConfiguration',
            private_link_ip_address,
            private_link_ip_allocation_method,
            private_link_primary,
            private_link_subnet_id
        )
        privateLinkConfigurations.append({
            'name': private_link_name,
            'properties': {
                'ipConfigurations': [default_private_link_ip_config]
            }
        })

    frontend_ip_configs = []

    # 4 combinations are valid for creating application gateway regarding to private IP and public IP
    # --------------------------------------------------------------------------------------------|
    # |                   |        private_ip_address         |        private_ip_address         |
    # |                   |         it not None               |          is None                  |
    # --------------------------------------------------------------------------------------------|
    # |                   |  private_ip_allocation: "Static"  | private_ip_allocation: "Dynamic"  |
    # |                   |  frontend_private_ip built: yes   | frontend_private_ip built: no     |
    # | public_ip_address |                                   |                                   |
    # |    it not None    | frontend_public_ip built: yes     | frontend_public_ip built: no      |
    # |                   |                                   |                                   |
    # |                   | 2 frontend IP configs entries     | 1 frontend IP configs entry       |
    # |                   |                                   |                                   |
    # |                   | frontend_ip_config_id: public_ip  | frontend_ip_config_id:public_ip   |
    # |                   |                                   |                                   |
    # |                   | private link link to public IP    | private link link to public IP    |
    # |-------------------------------------------------------------------------------------------|
    # |                   | private_ip_allocation: "Static"   | private_ip_allocation: "Dynamic"  |
    # | public_ip_address | frontend_private_ip built: yes    | frontend_private_ip built: no     |
    # |     is None       |                                   |                                   |
    # |                   | frontend_public_ip built: no      | frontend_public_ip built: no      |
    # |                   |                                   |                                   |
    # |                   | 1 frontend IP configs entry       | 1 frontend IP configs entry       |
    # |                   |                                   |                                   |
    # |                   | frontend_ip_config_id: priavte_ip | frontend_ip_config_id: priavte_ip |
    # |                   |                                   |                                   |
    # |                   | private link link to private IP   | private link link to private IP   |
    # |-------------------------------------------------------------------------------------------|
    if private_ip_address is not None or public_ip_id is None:
        enable_private_link = False if public_ip_id else enable_private_link
        frontend_private_ip = _build_frontend_ip_config(cmd, frontend_private_ip_name,
                                                        subnet_id=subnet_id,
                                                        private_ip_address=private_ip_address,
                                                        private_ip_allocation=private_ip_allocation,
                                                        enable_private_link=enable_private_link,
                                                        private_link_configuration_id=private_link_configuration_id)
        frontend_ip_configs.append(frontend_private_ip)

        frontend_ip_config_id = _ag_subresource_id('frontendIPConfigurations', frontend_private_ip_name)
    if public_ip_id:
        frontend_public_ip = _build_frontend_ip_config(cmd, frontend_public_ip_name,
                                                       public_ip_id=public_ip_id,
                                                       enable_private_link=enable_private_link,
                                                       private_link_configuration_id=private_link_configuration_id)
        frontend_ip_configs.append(frontend_public_ip)

        frontend_ip_config_id = _ag_subresource_id('frontendIPConfigurations', frontend_public_ip_name)

    http_listener = {
        'name': http_listener_name,
        'properties': {
            'FrontendIpConfiguration': {'Id': frontend_ip_config_id},
            'FrontendPort': {'Id': frontend_port_id},
            'Protocol': http_listener_protocol,
            'SslCertificate': None
        }
    }
    if cert_data:
        http_listener['properties'].update({'SslCertificate': {'id': ssl_cert_id}})
        ssl_cert = {
            'name': ssl_cert_name,
            'properties': {
                'data': cert_data,
            }
        }
        if cert_password:
            ssl_cert['properties']['password'] = "[parameters('certPassword')]"

    if key_vault_secret_id:
        http_listener['properties'].update({'SslCertificate': {'id': ssl_cert_id}})
        ssl_cert = {
            'name': ssl_cert_name,
            'properties': {
                'keyVaultSecretId': key_vault_secret_id,
            }
        }
    if ssl_profile_id and cmd.supported_api_version(min_api='2020-06-01'):
        http_listener['properties'].update({'sslProfile': {'id': ssl_profile_id}})

    backend_http_settings = {
        'name': http_settings_name,
        'properties': {
            'Port': http_settings_port,
            'Protocol': http_settings_protocol,
            'CookieBasedAffinity': cookie_based_affinity
        }
    }
    if cmd.supported_api_version(min_api='2016-12-01'):
        backend_http_settings['properties']['connectionDraining'] = {
            'enabled': bool(connection_draining_timeout),
            'drainTimeoutInSec': connection_draining_timeout if connection_draining_timeout else 1
        }

    ag_properties = {
        'backendAddressPools': [backend_address_pool],
        'backendHttpSettingsCollection': [backend_http_settings],
        'frontendIPConfigurations': frontend_ip_configs,
        'frontendPorts': [
            {
                'name': frontend_port_name,
                'properties': {
                    'Port': frontend_port
                }
            }
        ],
        'gatewayIPConfigurations': [
            {
                'name': frontend_public_ip_name if public_ip_id else frontend_private_ip_name,
                'properties': {
                    'subnet': {'id': subnet_id}
                }
            }
        ],
        'httpListeners': [http_listener],
        'sku': {
            'name': sku_name,
            'tier': sku_tier,
            'capacity': capacity
        },
        'requestRoutingRules': [
            {
                'Name': routing_rule_name,
                'properties': {
                    'RuleType': routing_rule_type,
                    'httpListener': {'id': http_listener_id},
                    'backendAddressPool': {'id': backend_address_pool_id},
                    'backendHttpSettings': {'id': backend_http_settings_id}
                }
            }
        ],
        'privateLinkConfigurations': privateLinkConfigurations,
    }
    if ssl_cert:
        ag_properties.update({'sslCertificates': [ssl_cert]})
    if enable_http2 and cmd.supported_api_version(min_api='2017-10-01'):
        ag_properties.update({'enableHttp2': enable_http2})
    if min_capacity and cmd.supported_api_version(min_api='2018-07-01'):
        if 'autoscaleConfiguration' not in ag_properties:
            ag_properties['autoscaleConfiguration'] = {}
        ag_properties['autoscaleConfiguration'].update({'minCapacity': min_capacity})
        ag_properties['sku'].pop('capacity', None)
    if max_capacity and cmd.supported_api_version(min_api='2018-12-01'):
        if 'autoscaleConfiguration' not in ag_properties:
            ag_properties['autoscaleConfiguration'] = {}
        ag_properties['autoscaleConfiguration'].update({'maxCapacity': max_capacity})
        ag_properties['sku'].pop('capacity', None)
    if custom_error_pages and cmd.supported_api_version(min_api='2018-08-01'):
        ag_properties.update({'customErrorConfigurations': custom_error_pages})
    if firewall_policy and cmd.supported_api_version(min_api='2018-12-01'):
        ag_properties.update({'firewallPolicy': {'id': firewall_policy}})

    # mutual authentication support
    if cmd.supported_api_version(min_api='2020-06-01') and trusted_client_certificates:
        parameters = []
        for item in trusted_client_certificates:
            parameters.append(
                {
                    "name": item['name'],
                    "properties": {
                        "data": item['data']
                    }
                }
            )
        ag_properties.update({"trustedClientCertificates": parameters})

    # ssl profiles
    if cmd.supported_api_version(min_api='2020-06-01') and ssl_profile:
        parameters = []
        for item in ssl_profile:
            parameter = {
                "name": item['name'],
                "properties": {
                    "sslPolicy": {}
                }
            }
            if 'policy_name' in item:
                parameter['properties']['sslPolicy'].update({"policyName": item['policy_name']})
            if 'policy_type' in item:
                parameter['properties']['sslPolicy'].update({"policyType": item['policy_type']})
            if 'min_protocol_version' in item:
                parameter['properties']['sslPolicy'].update({"minProtocolVersion": item['min_protocol_version']})
            if 'cipher_suites' in item:
                parameter['properties']['sslPolicy'].update({"cipherSuites": item['cipher_suites']})
            if 'client_auth_configuration' in item:
                parameter['properties'].update(
                    {"clientAuthConfiguration": {"verifyClientCertIssuerDN": item['client_auth_configuration']}})
            if 'trusted_client_certificates' in item:
                parameter['properties'].update(
                    {"trustedClientCertificates": [{"id": id['trusted_client_certificates']} for id in item]})

            parameters.append(parameter)

        ag_properties.update({"sslProfiles": parameters})

    ag = {
        'type': 'Microsoft.Network/applicationGateways',
        'name': name,
        'location': location,
        'tags': tags,
        'apiVersion': cmd.get_api_version(),
        'dependsOn': [],
        'properties': ag_properties
    }
    if cmd.supported_api_version(min_api='2018-08-01'):
        ag.update({'zones': zones})
    if user_assigned_identity and cmd.supported_api_version(min_api='2018-12-01'):
        ag.update(
            {
                "identity": {
                    "type": "UserAssigned",
                    "userAssignedIdentities": {
                        user_assigned_identity: {}
                    }
                }
            }
        )

    return ag


def build_load_balancer_resource(cmd, name, location, tags, backend_pool_name, frontend_ip_name, public_ip_id,
                                 subnet_id, private_ip_address, private_ip_allocation,
                                 sku, frontend_ip_zone, private_ip_address_version, tier=None,
                                 edge_zone=None, edge_zone_type=None):
    frontend_ip_config = _build_frontend_ip_config(cmd, frontend_ip_name, public_ip_id, subnet_id, private_ip_address,
                                                   private_ip_allocation, frontend_ip_zone, private_ip_address_version)

    lb_properties = {
        'backendAddressPools': [
            {
                'name': backend_pool_name
            }
        ],
        'frontendIPConfigurations': [frontend_ip_config]
    }

    # when sku is 'gateway', 'tunnelInterfaces' can't be None. Otherwise service will response error
    if cmd.supported_api_version(min_api='2021-02-01') and sku and str(sku).lower() == 'gateway':
        lb_properties['backendAddressPools'][0]['properties'] = {
            'tunnelInterfaces': [{'protocol': 'VXLAN',
                                  'type': 'Internal',
                                  "identifier": 900}]}

    lb = {
        'type': 'Microsoft.Network/loadBalancers',
        'name': name,
        'location': location,
        'tags': tags,
        'apiVersion': cmd.get_api_version(),
        'dependsOn': [],
        'properties': lb_properties
    }
    if sku and cmd.supported_api_version(min_api='2017-08-01'):
        lb['sku'] = {'name': sku}
    if tier and cmd.supported_api_version(min_api='2020-07-01'):
        lb['sku'].update({'tier': tier})
    if edge_zone and edge_zone_type:
        lb['extendedLocation'] = {'name': edge_zone, 'type': edge_zone_type}
    return lb


def build_public_ip_resource(cmd, name, location, tags, address_allocation, dns_name, sku, zone, tier=None,
                             edge_zone=None, edge_zone_type=None):
    public_ip_properties = {'publicIPAllocationMethod': address_allocation}

    if dns_name:
        public_ip_properties['dnsSettings'] = {'domainNameLabel': dns_name}

    public_ip = {
        'apiVersion': cmd.get_api_version(),
        'type': 'Microsoft.Network/publicIPAddresses',
        'name': name,
        'location': location,
        'tags': tags,
        'dependsOn': [],
        'properties': public_ip_properties
    }
    if sku and cmd.supported_api_version(min_api='2017-08-01'):
        public_ip['sku'] = {'name': sku}
    if tier and cmd.supported_api_version(min_api='2020-07-01'):
        if not sku:
            public_ip['sku'] = {'name': 'Basic'}
        public_ip['sku'].update({'tier': tier})
    if zone and cmd.supported_api_version(min_api='2017-06-01'):
        public_ip['zones'] = zone
    if edge_zone and edge_zone_type:
        public_ip['extendedLocation'] = {'name': edge_zone, 'type': edge_zone_type}
    return public_ip


def build_vnet_resource(_, name, location, tags, vnet_prefix=None, subnet=None, subnet_prefix=None, dns_servers=None,
                        enable_private_link=False, private_link_subnet=None, private_link_subnet_prefix=None):
    vnet = {
        'name': name,
        'type': 'Microsoft.Network/virtualNetworks',
        'location': location,
        'apiVersion': '2015-06-15',
        'dependsOn': [],
        'tags': tags,
        'properties': {
            'addressSpace': {'addressPrefixes': [vnet_prefix]},
            'subnets': []
        }
    }
    if dns_servers:
        vnet['properties']['dhcpOptions'] = {
            'dnsServers': dns_servers
        }
    if subnet:
        vnet['properties']['subnets'].append({
            'name': subnet,
            'properties': {
                'addressPrefix': subnet_prefix
            }
        })
    if enable_private_link:
        vnet['properties']['subnets'].append({
            'name': private_link_subnet,
            'properties': {
                'addressPrefix': private_link_subnet_prefix,
                'privateLinkServiceNetworkPolicies': 'Disabled',
            }
        })

    return vnet


def build_vpn_connection_resource(cmd, name, location, tags, gateway1, gateway2, vpn_type, authorization_key,
                                  enable_bgp, routing_weight, shared_key, use_policy_based_traffic_selectors,
                                  express_route_gateway_bypass, ingress_nat_rule, egress_nat_rule):
    vpn_properties = {
        'virtualNetworkGateway1': {'id': gateway1},
        'enableBgp': enable_bgp,
        'connectionType': vpn_type,
        'routingWeight': routing_weight
    }
    if authorization_key:
        vpn_properties['authorizationKey'] = "[parameters('authorizationKey')]"
    if cmd.supported_api_version(min_api='2017-03-01'):
        vpn_properties['usePolicyBasedTrafficSelectors'] = use_policy_based_traffic_selectors
    if cmd.supported_api_version(min_api='2018-07-01'):
        vpn_properties['expressRouteGatewayBypass'] = express_route_gateway_bypass

    # add scenario specific properties
    if shared_key:
        shared_key = "[parameters('sharedKey')]"
    if vpn_type == 'IPSec':
        vpn_properties.update({
            'localNetworkGateway2': {'id': gateway2},
            'sharedKey': shared_key
        })
    elif vpn_type == 'Vnet2Vnet':
        vpn_properties.update({
            'virtualNetworkGateway2': {'id': gateway2},
            'sharedKey': shared_key
        })
    elif vpn_type == 'ExpressRoute':
        vpn_properties.update({
            'peer': {'id': gateway2}
        })

    if ingress_nat_rule:
        vpn_properties['ingressNatRules'] = [{'id': rule} for rule in ingress_nat_rule]

    if egress_nat_rule:
        vpn_properties['egressNatRules'] = [{'id': rule} for rule in egress_nat_rule]

    vpn_connection = {
        'type': 'Microsoft.Network/connections',
        'name': name,
        'location': location,
        'tags': tags,
        'apiVersion': '2015-06-15',
        'dependsOn': [],
        'properties': vpn_properties if vpn_type != 'VpnClient' else {}
    }
    return vpn_connection
